﻿//Copyright © 2009 Max 'Shrinker' Wieden
//Refer to license.txt for license details

using System;
using System.Collections.Generic;
using System.Globalization;
using System.IO;
using VMFToMicrobrush.VMFParser;

namespace VMFToMicrobrush
{
	public class Vec3
	{
		public readonly float X, Y, Z;

		public Vec3(float x, float y, float z)
		{
			X = x;
			Y = y;
			Z = z;
		}

		public static Vec3 operator -(Vec3 a, Vec3 b)
		{
			return new Vec3(a.X-b.X, a.Y-b.Y, a.Z-b.Z);
		}
	}

	public class Face
	{
		public readonly Vec3 P0, U, V;

		public Face(Vec3 p0, Vec3 u, Vec3 v)
		{
			P0 = p0;
			U  = u ;
			V  = v ;
		}
	}

	public class Brush
	{
		public readonly Face[] Faces;

		public Brush(Face[] faces)
		{
			Faces = faces;
		}
	}

	public class Program
	{
		private VMFPlaneParser.Parser planeParser;
		//private CultureInfo culture;

		private static void Main(string[] args)
		{
			Program p = new Program();
			p.main(args);
		}

		private Face parseFace(string s)
		{
			VMFPlaneParser.RuleInstance plane = planeParser.Parse(s.Substring(1, s.Length-2));
			float x, y, z;
			x = float.Parse(((VMFPlaneParser.Token)plane.Children[1]).Image);
			y = float.Parse(((VMFPlaneParser.Token)plane.Children[2]).Image);
			z = float.Parse(((VMFPlaneParser.Token)plane.Children[3]).Image);
			Vec3 p0 = new Vec3(x, y, z);
			x = float.Parse(((VMFPlaneParser.Token)plane.Children[6]).Image);
			y = float.Parse(((VMFPlaneParser.Token)plane.Children[7]).Image);
			z = float.Parse(((VMFPlaneParser.Token)plane.Children[8]).Image);
			Vec3 p1 = new Vec3(x, y, z);
			x = float.Parse(((VMFPlaneParser.Token)plane.Children[11]).Image);
			y = float.Parse(((VMFPlaneParser.Token)plane.Children[12]).Image);
			z = float.Parse(((VMFPlaneParser.Token)plane.Children[13]).Image);
			Vec3 p2 = new Vec3(x, y, z);
			Vec3 u = p0-p1;
			Vec3 v = p2-p1;
			return new Face(p1, u, v);
		}

		private List<Brush> scanForBrushes(RuleInstance node)
		{
			List<Brush> result = new List<Brush>();
			if (node.Type == RuleType.Rule_Solid)
			{
				List<Face> faces = new List<Face>();
				foreach (Node child in node.Children)
					if (child.Type == RuleType.Rule_Side)
						for (int i = 2; i < ((RuleInstance)child).Children.Length; i += 2)
							if (((RuleInstance)child).Children[i].Type == RuleType.Tok_StringLit && ((Token)((RuleInstance)child).Children[i]).Image.Equals("\"plane\""))
								faces.Add(parseFace(((Token)((RuleInstance)child).Children[i+1]).Image));
				result.Add(new Brush(faces.ToArray()));
			}
			else
				foreach (Node child in node.Children)
					if (child.GetType() == typeof(RuleInstance))
						result.AddRange(scanForBrushes((RuleInstance)child));
			return result;
		}

		private byte[] convertBrushData(Brush[] brushes)
		{
			List<byte> result = new List<byte>();
			result.AddRange(BitConverter.GetBytes(brushes.Length));
			foreach (Brush b in brushes)
			{
				result.AddRange(BitConverter.GetBytes(b.Faces.Length));
				foreach (Face f in b.Faces)
				{
					result.AddRange(BitConverter.GetBytes( f.P0.X/8));
					result.AddRange(BitConverter.GetBytes( f.P0.Z/8));
					result.AddRange(BitConverter.GetBytes(-f.P0.Y/8));
					result.AddRange(BitConverter.GetBytes( f.U .X/8));
					result.AddRange(BitConverter.GetBytes( f.U .Z/8));
					result.AddRange(BitConverter.GetBytes(-f.U .Y/8));
					result.AddRange(BitConverter.GetBytes( f.V .X/8));
					result.AddRange(BitConverter.GetBytes( f.V .Z/8));
					result.AddRange(BitConverter.GetBytes(-f.V .Y/8));
				}
			}
			return result.ToArray();
		}

		private void main(string[] args)
		{
			string nl = Environment.NewLine;
			try
			{
				Console.WriteLine("== VMFToMicrobrush ==");
				if (args.Length < 2)
					throw new ArgumentException("usage: VMFToMicrobrush input.vmf output.dat");

				Console.WriteLine("Reading and parsing input file...");
				StreamReader r = File.OpenText(args[0]);
				RuleInstance root;
				try
				{
					VMFParser.Parser p = new VMFParser.Parser();
					root = p.Parse(r.ReadToEnd());
				}
				finally
				{
					r.Close();
				}

				Console.WriteLine("Converting...");
				planeParser = new VMFPlaneParser.Parser();
//				culture = CultureInfo.GetCultureInfo("");
				List<Brush> brushes = scanForBrushes(root);
				byte[] buf = convertBrushData(brushes.ToArray());

				Console.WriteLine("Saving output file...");
				FileStream w = File.Create(args[1]);
				try
				{
					w.Write(buf, 0, buf.Length);
				}
				finally
				{
					w.Close();
				}
			}
			catch (Exception e)
			{
				ConsoleColor color = Console.ForegroundColor;
				Console.ForegroundColor = ConsoleColor.Red;
				Console.WriteLine("An exception of type "+e.GetType()+" occurred, message: \""+e.Message+'\"');
				Console.WriteLine(e.StackTrace);
				Console.ForegroundColor = color;
			}
			Console.WriteLine("Done.");
		}
	}
}
