using System;
using System.Collections.Generic;

namespace VMFToMicrobrush.VMFParser
{
	public class ParserException: Exception
	{
		public ParserException(string message)
			:base(message)
		{
		}
	}

	public enum RuleType
	{
		Tok_EOF,
		Tok_LBrace,
		Tok_RBrace,
		Tok_Name,
			Tok_allowed_verts,
			Tok_alphas,
			Tok_camera,
			Tok_cameras,
			Tok_connections,
			Tok_cordon,
			Tok_dispinfo,
			Tok_distances,
			Tok_editor,
			Tok_entity,
			Tok_normals,
			Tok_offsets,
			Tok_offset_normals,
			Tok_side,
			Tok_solid,
			Tok_triangle_tags,
			Tok_versioninfo,
			Tok_viewsettings,
			Tok_visgroup,
			Tok_visgroups,
			Tok_world,
		Tok_StringLit,
		Tok_Whitespace,
		Rule_Start,
		Rule_VersionInfo,
		Rule_VisGroups,
		Rule_VisGroup,
		Rule_ViewSettings,
		Rule_World,
		Rule_Solid,
		Rule_Side,
		Rule_DispInfo,
		Rule_Normals,
		Rule_Distances,
		Rule_Offsets,
		Rule_Offset_Normals,
		Rule_Alphas,
		Rule_Triangle_Tags,
		Rule_Allowed_Verts,
		Rule_Editor,
		Rule_Entity,
		Rule_Connections,
		Rule_Cameras,
		Rule_Camera,
		Rule_Cordon
	}

	public abstract class Node
	{
		public readonly RuleType Type;

		public Node(RuleType type)
		{
			Type = type;
		}
	}

	public class RuleInstance: Node //represents an inner node
	{
		public readonly Node[] Children;

		public RuleInstance(RuleType type, Node[] children)
			:base(type)
		{
			Children = children;
		}
	}

	public class Token: Node //represents a leaf node
	{
		public readonly Token[] PrecedingFillerTokens;
		public readonly string  Image                ;

		private static Dictionary<string, RuleType> specForTok_Name;

		static Token()
		{
			specForTok_Name = new Dictionary<string, RuleType>();
			specForTok_Name["allowed_verts"] = RuleType.Tok_allowed_verts;
			specForTok_Name["alphas"] = RuleType.Tok_alphas;
			specForTok_Name["camera"] = RuleType.Tok_camera;
			specForTok_Name["cameras"] = RuleType.Tok_cameras;
			specForTok_Name["connections"] = RuleType.Tok_connections;
			specForTok_Name["cordon"] = RuleType.Tok_cordon;
			specForTok_Name["dispinfo"] = RuleType.Tok_dispinfo;
			specForTok_Name["distances"] = RuleType.Tok_distances;
			specForTok_Name["editor"] = RuleType.Tok_editor;
			specForTok_Name["entity"] = RuleType.Tok_entity;
			specForTok_Name["normals"] = RuleType.Tok_normals;
			specForTok_Name["offsets"] = RuleType.Tok_offsets;
			specForTok_Name["offset_normals"] = RuleType.Tok_offset_normals;
			specForTok_Name["side"] = RuleType.Tok_side;
			specForTok_Name["solid"] = RuleType.Tok_solid;
			specForTok_Name["triangle_tags"] = RuleType.Tok_triangle_tags;
			specForTok_Name["versioninfo"] = RuleType.Tok_versioninfo;
			specForTok_Name["viewsettings"] = RuleType.Tok_viewsettings;
			specForTok_Name["visgroup"] = RuleType.Tok_visgroup;
			specForTok_Name["visgroups"] = RuleType.Tok_visgroups;
			specForTok_Name["world"] = RuleType.Tok_world;
		}

		private static RuleType specializeType(RuleType type, string image)
		{
			RuleType newType;
			switch (type)
			{
				case RuleType.Tok_Name:
					if (specForTok_Name.TryGetValue(image, out newType))
						type = newType;
				break;
			}
			return type;
		}

		public Token(Token[] precedingFillerTokens, RuleType type, string image)
			:base(specializeType(type, image))
		{
			PrecedingFillerTokens = precedingFillerTokens;
			Image                 = image                ;
		}
	}

	public class Parser
	{
		private static Dictionary<RuleType, string> ruleNames;

		static Parser()
		{
			ruleNames = new Dictionary<RuleType,string>();
			ruleNames[RuleType.Tok_EOF] = "EOF";
			ruleNames[RuleType.Tok_LBrace] = "LBrace";
			ruleNames[RuleType.Tok_RBrace] = "RBrace";
			ruleNames[RuleType.Tok_Name] = "Name";
			ruleNames[RuleType.Tok_StringLit] = "StringLit";
			ruleNames[RuleType.Tok_Whitespace] = "Whitespace";
			ruleNames[RuleType.Rule_Start] = "Start";
			ruleNames[RuleType.Rule_VersionInfo] = "VersionInfo";
			ruleNames[RuleType.Rule_VisGroups] = "VisGroups";
			ruleNames[RuleType.Rule_VisGroup] = "VisGroup";
			ruleNames[RuleType.Rule_ViewSettings] = "ViewSettings";
			ruleNames[RuleType.Rule_World] = "World";
			ruleNames[RuleType.Rule_Solid] = "Solid";
			ruleNames[RuleType.Rule_Side] = "Side";
			ruleNames[RuleType.Rule_DispInfo] = "DispInfo";
			ruleNames[RuleType.Rule_Normals] = "Normals";
			ruleNames[RuleType.Rule_Distances] = "Distances";
			ruleNames[RuleType.Rule_Offsets] = "Offsets";
			ruleNames[RuleType.Rule_Offset_Normals] = "Offset_Normals";
			ruleNames[RuleType.Rule_Alphas] = "Alphas";
			ruleNames[RuleType.Rule_Triangle_Tags] = "Triangle_Tags";
			ruleNames[RuleType.Rule_Allowed_Verts] = "Allowed_Verts";
			ruleNames[RuleType.Rule_Editor] = "Editor";
			ruleNames[RuleType.Rule_Entity] = "Entity";
			ruleNames[RuleType.Rule_Connections] = "Connections";
			ruleNames[RuleType.Rule_Cameras] = "Cameras";
			ruleNames[RuleType.Rule_Camera] = "Camera";
			ruleNames[RuleType.Rule_Cordon] = "Cordon";
		}

		public static string GetRuleName(RuleType type)
		{
			return ruleNames[type];
		}

		private string source;
		private int index, length, laOffset;
		private List <Token> fetchedTokens;
		private Stack<bool > laSuccess    ;
		private Stack<int  > laOffsets    ;
		private Stack<bool > onceOrMoreB  ;

		public Parser()
		{
			fetchedTokens = new List <Token>();
			laSuccess     = new Stack<bool >();
			laOffsets     = new Stack<int  >();
			onceOrMoreB   = new Stack<bool >();
		}

		public RuleInstance Parse(string source)
		{
			this.source = source;
			index    =             0;
			length   = source.Length;
			laOffset =             0;
			RuleInstance r;
			try
			{
				r = matchStart();
			}
			finally
			{
				this.source = null;
				fetchedTokens.Clear();
				laSuccess    .Clear();
				laOffsets    .Clear();
				onceOrMoreB  .Clear();
			}
			return r;
		}

		private Token nextToken()
		{
			return nextToken(true);
		}

		private Token fetchToken(int offset)
		{
			while (fetchedTokens.Count <= offset)
			{
				Token tok = nextToken(false);
				fetchedTokens.Add(tok);
				//if token stream is exhausted, return EOF
				if (tok.Type == RuleType.Tok_EOF)
				{
					offset = fetchedTokens.Count-1;
					break;
				}
			}
			return fetchedTokens[offset];
		}

		private Token[] getFillerTokens()
		{
			List<Token> result = new List<Token>();
			Stack<int> indices = new Stack<int>();
			char c;
			bool pass;
			while (true)
			{
				int oldIndex = index;
				bool passedThisIteration = false;

				//Token "Whitespace"
				pass = true;
				indices.Push(index);
				onceOrMoreB.Push(false);
				while (true)
				{
					indices.Push(index);
					if (index < length && ((c = source[index]) == '\t' || c == ' ' || c == '\r' || c == '\n'))
						index++;
					else
						pass = false;
					if (pass)
					{
						onceOrMoreB.Pop();
						onceOrMoreB.Push(true);
						indices.Pop();
					}
					else
					{
						index = indices.Pop();
						break;
					}
				}
				pass = onceOrMoreB.Pop();
				if (pass)
					indices.Pop();
				else
					index = indices.Pop();
				if (pass)
				{
					result.Add(new Token(null, RuleType.Tok_Whitespace, source.Substring(oldIndex, index-oldIndex)));
					continue;
				}
				passedThisIteration |= pass;

				if (!passedThisIteration)
					break;
			}

			return result.ToArray();
		}

		private Token nextToken(bool useFetched)
		{
			if (useFetched && fetchedTokens.Count != 0)
			{
				Token tok = fetchedTokens[0];
				fetchedTokens.RemoveAt(0);
				return tok;
			}

			Token[] fillers = getFillerTokens();

			if (index == length)
				return new Token(fillers, RuleType.Tok_EOF, "");

			Stack<int> indices = new Stack<int>();
			char c;
			bool pass;
			int oldIndex = index;
			//Token "LBrace"
			pass = true;
			if (index < length && source[index] == '{')
				index++;
			else
				pass = false;
			if (pass)
				return new Token(fillers, RuleType.Tok_LBrace, source.Substring(oldIndex, index-oldIndex));

			//Token "RBrace"
			pass = true;
			if (index < length && source[index] == '}')
				index++;
			else
				pass = false;
			if (pass)
				return new Token(fillers, RuleType.Tok_RBrace, source.Substring(oldIndex, index-oldIndex));

			//Token "Name"
			pass = true;
			indices.Push(index);
			onceOrMoreB.Push(false);
			while (true)
			{
				indices.Push(index);
				indices.Push(index);
				if (index < length && (c = source[index]) >= 'a' && c <= 'z')
					index++;
				else
					pass = false;
				if (!pass)
				{
					pass = true;
					index = indices.Peek();
					if (index < length && source[index] == '_')
						index++;
					else
						pass = false;
				}
				indices.Pop();
				if (pass)
				{
					onceOrMoreB.Pop();
					onceOrMoreB.Push(true);
					indices.Pop();
				}
				else
				{
					index = indices.Pop();
					break;
				}
			}
			pass = onceOrMoreB.Pop();
			if (pass)
				indices.Pop();
			else
				index = indices.Pop();
			if (pass)
				return new Token(fillers, RuleType.Tok_Name, source.Substring(oldIndex, index-oldIndex));

			//Token "StringLit"
			pass = true;
			if (index < length && source[index] == '\"')
				index++;
			else
				pass = false;
			if (pass)
			{
				while (true)
				{
					indices.Push(index);
					if (index < length && source[index] != '\"')
						index++;
					else
						pass = false;
					if (pass)
						indices.Pop();
					else
					{
						pass = true;
						index = indices.Pop();
						break;
					}
				}
			}
			if (pass)
			{
				if (index < length && source[index] == '\"')
					index++;
				else
					pass = false;
			}
			if (pass)
				return new Token(fillers, RuleType.Tok_StringLit, source.Substring(oldIndex, index-oldIndex));

			throw new ParserException("No token match");
		}

		private RuleInstance matchStart()
		{
			List<Node> nodes = new List<Node>();
			Token tok;

			while (true)
			{
				if ((tok = fetchToken(laOffset)).Type == RuleType.Tok_cameras || tok.Type == RuleType.Tok_cordon || tok.Type == RuleType.Tok_entity || tok.Type == RuleType.Tok_versioninfo || tok.Type == RuleType.Tok_viewsettings || tok.Type == RuleType.Tok_visgroups || tok.Type == RuleType.Tok_world)
				{
					if (fetchToken(laOffset).Type == RuleType.Tok_versioninfo)
					{
						nodes.Add(matchVersionInfo());
					}
					else
					{
						if (fetchToken(laOffset).Type == RuleType.Tok_visgroups)
						{
							nodes.Add(matchVisGroups());
						}
						else
						{
							if (fetchToken(laOffset).Type == RuleType.Tok_viewsettings)
							{
								nodes.Add(matchViewSettings());
							}
							else
							{
								if (fetchToken(laOffset).Type == RuleType.Tok_world)
								{
									nodes.Add(matchWorld());
								}
								else
								{
									if (fetchToken(laOffset).Type == RuleType.Tok_entity)
									{
										nodes.Add(matchEntity());
									}
									else
									{
										if (fetchToken(laOffset).Type == RuleType.Tok_cameras)
										{
											nodes.Add(matchCameras());
										}
										else
										{
											nodes.Add(matchCordon());
										}
									}
								}
							}
						}
					}
				}
				else
					break;
			}

			return new RuleInstance(RuleType.Rule_Start, nodes.ToArray());
		}

		private RuleInstance matchVersionInfo()
		{
			List<Node> nodes = new List<Node>();
			Token tok;

			if ((tok = nextToken()).Type != RuleType.Tok_versioninfo)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);
			if ((tok = nextToken()).Type != RuleType.Tok_LBrace)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);
			while (true)
			{
				if (fetchToken(laOffset).Type == RuleType.Tok_StringLit)
				{
					if ((tok = nextToken()).Type != RuleType.Tok_StringLit)
						throw new ParserException("ERROR TODO");
					nodes.Add(tok);
					if ((tok = nextToken()).Type != RuleType.Tok_StringLit)
						throw new ParserException("ERROR TODO");
					nodes.Add(tok);
				}
				else
					break;
			}
			if ((tok = nextToken()).Type != RuleType.Tok_RBrace)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);

			return new RuleInstance(RuleType.Rule_VersionInfo, nodes.ToArray());
		}

		private RuleInstance matchVisGroups()
		{
			List<Node> nodes = new List<Node>();
			Token tok;

			if ((tok = nextToken()).Type != RuleType.Tok_visgroups)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);
			if ((tok = nextToken()).Type != RuleType.Tok_LBrace)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);
			while (true)
			{
				if (fetchToken(laOffset).Type == RuleType.Tok_visgroup)
				{
					nodes.Add(matchVisGroup());
				}
				else
					break;
			}
			if ((tok = nextToken()).Type != RuleType.Tok_RBrace)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);

			return new RuleInstance(RuleType.Rule_VisGroups, nodes.ToArray());
		}

		private RuleInstance matchVisGroup()
		{
			List<Node> nodes = new List<Node>();
			Token tok;

			if ((tok = nextToken()).Type != RuleType.Tok_visgroup)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);
			if ((tok = nextToken()).Type != RuleType.Tok_LBrace)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);
			while (true)
			{
				if (fetchToken(laOffset).Type == RuleType.Tok_StringLit)
				{
					if ((tok = nextToken()).Type != RuleType.Tok_StringLit)
						throw new ParserException("ERROR TODO");
					nodes.Add(tok);
					if ((tok = nextToken()).Type != RuleType.Tok_StringLit)
						throw new ParserException("ERROR TODO");
					nodes.Add(tok);
				}
				else
					break;
			}
			if ((tok = nextToken()).Type != RuleType.Tok_RBrace)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);

			return new RuleInstance(RuleType.Rule_VisGroup, nodes.ToArray());
		}

		private RuleInstance matchViewSettings()
		{
			List<Node> nodes = new List<Node>();
			Token tok;

			if ((tok = nextToken()).Type != RuleType.Tok_viewsettings)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);
			if ((tok = nextToken()).Type != RuleType.Tok_LBrace)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);
			while (true)
			{
				if (fetchToken(laOffset).Type == RuleType.Tok_StringLit)
				{
					if ((tok = nextToken()).Type != RuleType.Tok_StringLit)
						throw new ParserException("ERROR TODO");
					nodes.Add(tok);
					if ((tok = nextToken()).Type != RuleType.Tok_StringLit)
						throw new ParserException("ERROR TODO");
					nodes.Add(tok);
				}
				else
					break;
			}
			if ((tok = nextToken()).Type != RuleType.Tok_RBrace)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);

			return new RuleInstance(RuleType.Rule_ViewSettings, nodes.ToArray());
		}

		private RuleInstance matchWorld()
		{
			List<Node> nodes = new List<Node>();
			Token tok;

			if ((tok = nextToken()).Type != RuleType.Tok_world)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);
			if ((tok = nextToken()).Type != RuleType.Tok_LBrace)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);
			while (true)
			{
				if (fetchToken(laOffset).Type == RuleType.Tok_StringLit)
				{
					if ((tok = nextToken()).Type != RuleType.Tok_StringLit)
						throw new ParserException("ERROR TODO");
					nodes.Add(tok);
					if ((tok = nextToken()).Type != RuleType.Tok_StringLit)
						throw new ParserException("ERROR TODO");
					nodes.Add(tok);
				}
				else
					break;
			}
			while (true)
			{
				if (fetchToken(laOffset).Type == RuleType.Tok_solid)
				{
					nodes.Add(matchSolid());
				}
				else
					break;
			}
			if ((tok = nextToken()).Type != RuleType.Tok_RBrace)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);

			return new RuleInstance(RuleType.Rule_World, nodes.ToArray());
		}

		private RuleInstance matchSolid()
		{
			List<Node> nodes = new List<Node>();
			Token tok;

			if ((tok = nextToken()).Type != RuleType.Tok_solid)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);
			if ((tok = nextToken()).Type != RuleType.Tok_LBrace)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);
			if ((tok = nextToken()).Type != RuleType.Tok_StringLit)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);
			if ((tok = nextToken()).Type != RuleType.Tok_StringLit)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);
			while (true)
			{
				if (fetchToken(laOffset).Type == RuleType.Tok_side)
				{
					nodes.Add(matchSide());
				}
				else
					break;
			}
			if (fetchToken(laOffset).Type == RuleType.Tok_editor)
			{
				nodes.Add(matchEditor());
			}
			if ((tok = nextToken()).Type != RuleType.Tok_RBrace)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);

			return new RuleInstance(RuleType.Rule_Solid, nodes.ToArray());
		}

		private RuleInstance matchSide()
		{
			List<Node> nodes = new List<Node>();
			Token tok;

			if ((tok = nextToken()).Type != RuleType.Tok_side)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);
			if ((tok = nextToken()).Type != RuleType.Tok_LBrace)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);
			while (true)
			{
				if (fetchToken(laOffset).Type == RuleType.Tok_StringLit)
				{
					if ((tok = nextToken()).Type != RuleType.Tok_StringLit)
						throw new ParserException("ERROR TODO");
					nodes.Add(tok);
					if ((tok = nextToken()).Type != RuleType.Tok_StringLit)
						throw new ParserException("ERROR TODO");
					nodes.Add(tok);
				}
				else
					break;
			}
			if (fetchToken(laOffset).Type == RuleType.Tok_dispinfo)
			{
				nodes.Add(matchDispInfo());
			}
			if ((tok = nextToken()).Type != RuleType.Tok_RBrace)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);

			return new RuleInstance(RuleType.Rule_Side, nodes.ToArray());
		}

		private RuleInstance matchDispInfo()
		{
			List<Node> nodes = new List<Node>();
			Token tok;

			if ((tok = nextToken()).Type != RuleType.Tok_dispinfo)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);
			if ((tok = nextToken()).Type != RuleType.Tok_LBrace)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);
			while (true)
			{
				if (fetchToken(laOffset).Type == RuleType.Tok_StringLit)
				{
					if ((tok = nextToken()).Type != RuleType.Tok_StringLit)
						throw new ParserException("ERROR TODO");
					nodes.Add(tok);
					if ((tok = nextToken()).Type != RuleType.Tok_StringLit)
						throw new ParserException("ERROR TODO");
					nodes.Add(tok);
				}
				else
					break;
			}
			nodes.Add(matchNormals());
			nodes.Add(matchDistances());
			nodes.Add(matchOffsets());
			nodes.Add(matchOffset_Normals());
			nodes.Add(matchAlphas());
			nodes.Add(matchTriangle_Tags());
			nodes.Add(matchAllowed_Verts());
			if ((tok = nextToken()).Type != RuleType.Tok_RBrace)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);

			return new RuleInstance(RuleType.Rule_DispInfo, nodes.ToArray());
		}

		private RuleInstance matchNormals()
		{
			List<Node> nodes = new List<Node>();
			Token tok;

			if ((tok = nextToken()).Type != RuleType.Tok_normals)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);
			if ((tok = nextToken()).Type != RuleType.Tok_LBrace)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);
			while (true)
			{
				if (fetchToken(laOffset).Type == RuleType.Tok_StringLit)
				{
					if ((tok = nextToken()).Type != RuleType.Tok_StringLit)
						throw new ParserException("ERROR TODO");
					nodes.Add(tok);
					if ((tok = nextToken()).Type != RuleType.Tok_StringLit)
						throw new ParserException("ERROR TODO");
					nodes.Add(tok);
				}
				else
					break;
			}
			if ((tok = nextToken()).Type != RuleType.Tok_RBrace)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);

			return new RuleInstance(RuleType.Rule_Normals, nodes.ToArray());
		}

		private RuleInstance matchDistances()
		{
			List<Node> nodes = new List<Node>();
			Token tok;

			if ((tok = nextToken()).Type != RuleType.Tok_distances)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);
			if ((tok = nextToken()).Type != RuleType.Tok_LBrace)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);
			while (true)
			{
				if (fetchToken(laOffset).Type == RuleType.Tok_StringLit)
				{
					if ((tok = nextToken()).Type != RuleType.Tok_StringLit)
						throw new ParserException("ERROR TODO");
					nodes.Add(tok);
					if ((tok = nextToken()).Type != RuleType.Tok_StringLit)
						throw new ParserException("ERROR TODO");
					nodes.Add(tok);
				}
				else
					break;
			}
			if ((tok = nextToken()).Type != RuleType.Tok_RBrace)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);

			return new RuleInstance(RuleType.Rule_Distances, nodes.ToArray());
		}

		private RuleInstance matchOffsets()
		{
			List<Node> nodes = new List<Node>();
			Token tok;

			if ((tok = nextToken()).Type != RuleType.Tok_offsets)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);
			if ((tok = nextToken()).Type != RuleType.Tok_LBrace)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);
			while (true)
			{
				if (fetchToken(laOffset).Type == RuleType.Tok_StringLit)
				{
					if ((tok = nextToken()).Type != RuleType.Tok_StringLit)
						throw new ParserException("ERROR TODO");
					nodes.Add(tok);
					if ((tok = nextToken()).Type != RuleType.Tok_StringLit)
						throw new ParserException("ERROR TODO");
					nodes.Add(tok);
				}
				else
					break;
			}
			if ((tok = nextToken()).Type != RuleType.Tok_RBrace)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);

			return new RuleInstance(RuleType.Rule_Offsets, nodes.ToArray());
		}

		private RuleInstance matchOffset_Normals()
		{
			List<Node> nodes = new List<Node>();
			Token tok;

			if ((tok = nextToken()).Type != RuleType.Tok_offset_normals)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);
			if ((tok = nextToken()).Type != RuleType.Tok_LBrace)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);
			while (true)
			{
				if (fetchToken(laOffset).Type == RuleType.Tok_StringLit)
				{
					if ((tok = nextToken()).Type != RuleType.Tok_StringLit)
						throw new ParserException("ERROR TODO");
					nodes.Add(tok);
					if ((tok = nextToken()).Type != RuleType.Tok_StringLit)
						throw new ParserException("ERROR TODO");
					nodes.Add(tok);
				}
				else
					break;
			}
			if ((tok = nextToken()).Type != RuleType.Tok_RBrace)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);

			return new RuleInstance(RuleType.Rule_Offset_Normals, nodes.ToArray());
		}

		private RuleInstance matchAlphas()
		{
			List<Node> nodes = new List<Node>();
			Token tok;

			if ((tok = nextToken()).Type != RuleType.Tok_alphas)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);
			if ((tok = nextToken()).Type != RuleType.Tok_LBrace)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);
			while (true)
			{
				if (fetchToken(laOffset).Type == RuleType.Tok_StringLit)
				{
					if ((tok = nextToken()).Type != RuleType.Tok_StringLit)
						throw new ParserException("ERROR TODO");
					nodes.Add(tok);
					if ((tok = nextToken()).Type != RuleType.Tok_StringLit)
						throw new ParserException("ERROR TODO");
					nodes.Add(tok);
				}
				else
					break;
			}
			if ((tok = nextToken()).Type != RuleType.Tok_RBrace)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);

			return new RuleInstance(RuleType.Rule_Alphas, nodes.ToArray());
		}

		private RuleInstance matchTriangle_Tags()
		{
			List<Node> nodes = new List<Node>();
			Token tok;

			if ((tok = nextToken()).Type != RuleType.Tok_triangle_tags)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);
			if ((tok = nextToken()).Type != RuleType.Tok_LBrace)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);
			while (true)
			{
				if (fetchToken(laOffset).Type == RuleType.Tok_StringLit)
				{
					if ((tok = nextToken()).Type != RuleType.Tok_StringLit)
						throw new ParserException("ERROR TODO");
					nodes.Add(tok);
					if ((tok = nextToken()).Type != RuleType.Tok_StringLit)
						throw new ParserException("ERROR TODO");
					nodes.Add(tok);
				}
				else
					break;
			}
			if ((tok = nextToken()).Type != RuleType.Tok_RBrace)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);

			return new RuleInstance(RuleType.Rule_Triangle_Tags, nodes.ToArray());
		}

		private RuleInstance matchAllowed_Verts()
		{
			List<Node> nodes = new List<Node>();
			Token tok;

			if ((tok = nextToken()).Type != RuleType.Tok_allowed_verts)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);
			if ((tok = nextToken()).Type != RuleType.Tok_LBrace)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);
			while (true)
			{
				if (fetchToken(laOffset).Type == RuleType.Tok_StringLit)
				{
					if ((tok = nextToken()).Type != RuleType.Tok_StringLit)
						throw new ParserException("ERROR TODO");
					nodes.Add(tok);
					if ((tok = nextToken()).Type != RuleType.Tok_StringLit)
						throw new ParserException("ERROR TODO");
					nodes.Add(tok);
				}
				else
					break;
			}
			if ((tok = nextToken()).Type != RuleType.Tok_RBrace)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);

			return new RuleInstance(RuleType.Rule_Allowed_Verts, nodes.ToArray());
		}

		private RuleInstance matchEditor()
		{
			List<Node> nodes = new List<Node>();
			Token tok;

			if ((tok = nextToken()).Type != RuleType.Tok_editor)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);
			if ((tok = nextToken()).Type != RuleType.Tok_LBrace)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);
			while (true)
			{
				if (fetchToken(laOffset).Type == RuleType.Tok_StringLit)
				{
					if ((tok = nextToken()).Type != RuleType.Tok_StringLit)
						throw new ParserException("ERROR TODO");
					nodes.Add(tok);
					if ((tok = nextToken()).Type != RuleType.Tok_StringLit)
						throw new ParserException("ERROR TODO");
					nodes.Add(tok);
				}
				else
					break;
			}
			if ((tok = nextToken()).Type != RuleType.Tok_RBrace)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);

			return new RuleInstance(RuleType.Rule_Editor, nodes.ToArray());
		}

		private RuleInstance matchEntity()
		{
			List<Node> nodes = new List<Node>();
			Token tok;

			if ((tok = nextToken()).Type != RuleType.Tok_entity)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);
			if ((tok = nextToken()).Type != RuleType.Tok_LBrace)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);
			while (true)
			{
				if (fetchToken(laOffset).Type == RuleType.Tok_StringLit)
				{
					if ((tok = nextToken()).Type != RuleType.Tok_StringLit)
						throw new ParserException("ERROR TODO");
					nodes.Add(tok);
					if ((tok = nextToken()).Type != RuleType.Tok_StringLit)
						throw new ParserException("ERROR TODO");
					nodes.Add(tok);
				}
				else
					break;
			}
			if (fetchToken(laOffset).Type == RuleType.Tok_connections)
			{
				nodes.Add(matchConnections());
			}
			while (true)
			{
				if (fetchToken(laOffset).Type == RuleType.Tok_solid)
				{
					nodes.Add(matchSolid());
				}
				else
					break;
			}
			if (fetchToken(laOffset).Type == RuleType.Tok_editor)
			{
				nodes.Add(matchEditor());
			}
			if ((tok = nextToken()).Type != RuleType.Tok_RBrace)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);

			return new RuleInstance(RuleType.Rule_Entity, nodes.ToArray());
		}

		private RuleInstance matchConnections()
		{
			List<Node> nodes = new List<Node>();
			Token tok;

			if ((tok = nextToken()).Type != RuleType.Tok_connections)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);
			if ((tok = nextToken()).Type != RuleType.Tok_LBrace)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);
			while (true)
			{
				if (fetchToken(laOffset).Type == RuleType.Tok_StringLit)
				{
					if ((tok = nextToken()).Type != RuleType.Tok_StringLit)
						throw new ParserException("ERROR TODO");
					nodes.Add(tok);
					if ((tok = nextToken()).Type != RuleType.Tok_StringLit)
						throw new ParserException("ERROR TODO");
					nodes.Add(tok);
				}
				else
					break;
			}
			if ((tok = nextToken()).Type != RuleType.Tok_RBrace)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);

			return new RuleInstance(RuleType.Rule_Connections, nodes.ToArray());
		}

		private RuleInstance matchCameras()
		{
			List<Node> nodes = new List<Node>();
			Token tok;

			if ((tok = nextToken()).Type != RuleType.Tok_cameras)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);
			if ((tok = nextToken()).Type != RuleType.Tok_LBrace)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);
			while (true)
			{
				if (fetchToken(laOffset).Type == RuleType.Tok_StringLit)
				{
					if ((tok = nextToken()).Type != RuleType.Tok_StringLit)
						throw new ParserException("ERROR TODO");
					nodes.Add(tok);
					if ((tok = nextToken()).Type != RuleType.Tok_StringLit)
						throw new ParserException("ERROR TODO");
					nodes.Add(tok);
				}
				else
					break;
			}
			while (true)
			{
				if (fetchToken(laOffset).Type == RuleType.Tok_camera)
				{
					nodes.Add(matchCamera());
				}
				else
					break;
			}
			if ((tok = nextToken()).Type != RuleType.Tok_RBrace)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);

			return new RuleInstance(RuleType.Rule_Cameras, nodes.ToArray());
		}

		private RuleInstance matchCamera()
		{
			List<Node> nodes = new List<Node>();
			Token tok;

			if ((tok = nextToken()).Type != RuleType.Tok_camera)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);
			if ((tok = nextToken()).Type != RuleType.Tok_LBrace)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);
			while (true)
			{
				if (fetchToken(laOffset).Type == RuleType.Tok_StringLit)
				{
					if ((tok = nextToken()).Type != RuleType.Tok_StringLit)
						throw new ParserException("ERROR TODO");
					nodes.Add(tok);
					if ((tok = nextToken()).Type != RuleType.Tok_StringLit)
						throw new ParserException("ERROR TODO");
					nodes.Add(tok);
				}
				else
					break;
			}
			if ((tok = nextToken()).Type != RuleType.Tok_RBrace)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);

			return new RuleInstance(RuleType.Rule_Camera, nodes.ToArray());
		}

		private RuleInstance matchCordon()
		{
			List<Node> nodes = new List<Node>();
			Token tok;

			if ((tok = nextToken()).Type != RuleType.Tok_cordon)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);
			if ((tok = nextToken()).Type != RuleType.Tok_LBrace)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);
			while (true)
			{
				if (fetchToken(laOffset).Type == RuleType.Tok_StringLit)
				{
					if ((tok = nextToken()).Type != RuleType.Tok_StringLit)
						throw new ParserException("ERROR TODO");
					nodes.Add(tok);
					if ((tok = nextToken()).Type != RuleType.Tok_StringLit)
						throw new ParserException("ERROR TODO");
					nodes.Add(tok);
				}
				else
					break;
			}
			if ((tok = nextToken()).Type != RuleType.Tok_RBrace)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);

			return new RuleInstance(RuleType.Rule_Cordon, nodes.ToArray());
		}
	}
}
