using System;
using System.Collections.Generic;

namespace VMFToMicrobrush.VMFPlaneParser
{
	public class ParserException: Exception
	{
		public ParserException(string message)
			:base(message)
		{
		}
	}

	public enum RuleType
	{
		Tok_EOF,
		Tok_LPar,
		Tok_RPar,
		Tok_NumLit,
		Tok_Space,
		Rule_Start
	}

	public abstract class Node
	{
		public readonly RuleType Type;

		public Node(RuleType type)
		{
			Type = type;
		}
	}

	public class RuleInstance: Node //represents an inner node
	{
		public readonly Node[] Children;

		public RuleInstance(RuleType type, Node[] children)
			:base(type)
		{
			Children = children;
		}
	}

	public class Token: Node //represents a leaf node
	{
		public readonly Token[] PrecedingFillerTokens;
		public readonly string  Image                ;

		public Token(Token[] precedingFillerTokens, RuleType type, string image)
			:base(type)
		{
			PrecedingFillerTokens = precedingFillerTokens;
			Image                 = image                ;
		}
	}

	public class Parser
	{
		private static Dictionary<RuleType, string> ruleNames;

		static Parser()
		{
			ruleNames = new Dictionary<RuleType,string>();
			ruleNames[RuleType.Tok_EOF] = "EOF";
			ruleNames[RuleType.Tok_LPar] = "LPar";
			ruleNames[RuleType.Tok_RPar] = "RPar";
			ruleNames[RuleType.Tok_NumLit] = "NumLit";
			ruleNames[RuleType.Tok_Space] = "Space";
			ruleNames[RuleType.Rule_Start] = "Start";
		}

		public static string GetRuleName(RuleType type)
		{
			return ruleNames[type];
		}

		private string source;
		private int index, length, laOffset;
		private List <Token> fetchedTokens;
		private Stack<bool > laSuccess    ;
		private Stack<int  > laOffsets    ;
		private Stack<bool > onceOrMoreB  ;

		public Parser()
		{
			fetchedTokens = new List <Token>();
			laSuccess     = new Stack<bool >();
			laOffsets     = new Stack<int  >();
			onceOrMoreB   = new Stack<bool >();
		}

		public RuleInstance Parse(string source)
		{
			this.source = source;
			index    =             0;
			length   = source.Length;
			laOffset =             0;
			RuleInstance r;
			try
			{
				r = matchStart();
			}
			finally
			{
				this.source = null;
				fetchedTokens.Clear();
				laSuccess    .Clear();
				laOffsets    .Clear();
				onceOrMoreB  .Clear();
			}
			return r;
		}

		private Token nextToken()
		{
			return nextToken(true);
		}

		private Token fetchToken(int offset)
		{
			while (fetchedTokens.Count <= offset)
			{
				Token tok = nextToken(false);
				fetchedTokens.Add(tok);
				//if token stream is exhausted, return EOF
				if (tok.Type == RuleType.Tok_EOF)
				{
					offset = fetchedTokens.Count-1;
					break;
				}
			}
			return fetchedTokens[offset];
		}

		private Token[] getFillerTokens()
		{
			List<Token> result = new List<Token>();
			Stack<int> indices = new Stack<int>();
			char c;
			bool pass;
			while (true)
			{
				int oldIndex = index;
				bool passedThisIteration = false;

				//Token "Space"
				pass = true;
				if (index < length && source[index] == ' ')
					index++;
				else
					pass = false;
				if (pass)
				{
					result.Add(new Token(null, RuleType.Tok_Space, source.Substring(oldIndex, index-oldIndex)));
					continue;
				}
				passedThisIteration |= pass;

				if (!passedThisIteration)
					break;
			}

			return result.ToArray();
		}

		private Token nextToken(bool useFetched)
		{
			if (useFetched && fetchedTokens.Count != 0)
			{
				Token tok = fetchedTokens[0];
				fetchedTokens.RemoveAt(0);
				return tok;
			}

			Token[] fillers = getFillerTokens();

			if (index == length)
				return new Token(fillers, RuleType.Tok_EOF, "");

			Stack<int> indices = new Stack<int>();
			char c;
			bool pass;
			int oldIndex = index;
			//Token "LPar"
			pass = true;
			if (index < length && source[index] == '(')
				index++;
			else
				pass = false;
			if (pass)
				return new Token(fillers, RuleType.Tok_LPar, source.Substring(oldIndex, index-oldIndex));

			//Token "RPar"
			pass = true;
			if (index < length && source[index] == ')')
				index++;
			else
				pass = false;
			if (pass)
				return new Token(fillers, RuleType.Tok_RPar, source.Substring(oldIndex, index-oldIndex));

			//Token "NumLit"
			pass = true;
			indices.Push(index);
			if (index < length && source[index] == '-')
				index++;
			else
				pass = false;
			if (pass)
				indices.Pop();
			else
			{
				pass = true;
				index = indices.Pop();
			}
			if (pass)
			{
				indices.Push(index);
				indices.Push(index);
				onceOrMoreB.Push(false);
				while (true)
				{
					indices.Push(index);
					if (index < length && (c = source[index]) >= '0' && c <= '9')
						index++;
					else
						pass = false;
					if (pass)
					{
						onceOrMoreB.Pop();
						onceOrMoreB.Push(true);
						indices.Pop();
					}
					else
					{
						index = indices.Pop();
						break;
					}
				}
				pass = onceOrMoreB.Pop();
				if (pass)
					indices.Pop();
				else
					index = indices.Pop();
				if (pass)
				{
					indices.Push(index);
					if (index < length && source[index] == '.')
						index++;
					else
						pass = false;
					if (pass)
					{
						while (true)
						{
							indices.Push(index);
							if (index < length && (c = source[index]) >= '0' && c <= '9')
								index++;
							else
								pass = false;
							if (pass)
								indices.Pop();
							else
							{
								pass = true;
								index = indices.Pop();
								break;
							}
						}
					}
					if (pass)
						indices.Pop();
					else
					{
						pass = true;
						index = indices.Pop();
					}
				}
				if (!pass)
				{
					pass = true;
					index = indices.Peek();
					if (index < length && source[index] == '.')
						index++;
					else
						pass = false;
					if (pass)
					{
						indices.Push(index);
						onceOrMoreB.Push(false);
						while (true)
						{
							indices.Push(index);
							if (index < length && (c = source[index]) >= '0' && c <= '9')
								index++;
							else
								pass = false;
							if (pass)
							{
								onceOrMoreB.Pop();
								onceOrMoreB.Push(true);
								indices.Pop();
							}
							else
							{
								index = indices.Pop();
								break;
							}
						}
						pass = onceOrMoreB.Pop();
						if (pass)
							indices.Pop();
						else
							index = indices.Pop();
					}
				}
				indices.Pop();
			}
			if (pass)
				return new Token(fillers, RuleType.Tok_NumLit, source.Substring(oldIndex, index-oldIndex));

			throw new ParserException("No token match");
		}

		private RuleInstance matchStart()
		{
			List<Node> nodes = new List<Node>();
			Token tok;

			if ((tok = nextToken()).Type != RuleType.Tok_LPar)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);
			if ((tok = nextToken()).Type != RuleType.Tok_NumLit)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);
			if ((tok = nextToken()).Type != RuleType.Tok_NumLit)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);
			if ((tok = nextToken()).Type != RuleType.Tok_NumLit)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);
			if ((tok = nextToken()).Type != RuleType.Tok_RPar)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);
			if ((tok = nextToken()).Type != RuleType.Tok_LPar)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);
			if ((tok = nextToken()).Type != RuleType.Tok_NumLit)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);
			if ((tok = nextToken()).Type != RuleType.Tok_NumLit)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);
			if ((tok = nextToken()).Type != RuleType.Tok_NumLit)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);
			if ((tok = nextToken()).Type != RuleType.Tok_RPar)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);
			if ((tok = nextToken()).Type != RuleType.Tok_LPar)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);
			if ((tok = nextToken()).Type != RuleType.Tok_NumLit)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);
			if ((tok = nextToken()).Type != RuleType.Tok_NumLit)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);
			if ((tok = nextToken()).Type != RuleType.Tok_NumLit)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);
			if ((tok = nextToken()).Type != RuleType.Tok_RPar)
				throw new ParserException("ERROR TODO");
			nodes.Add(tok);

			return new RuleInstance(RuleType.Rule_Start, nodes.ToArray());
		}
	}
}
